﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection.Emit;
using System.Collections;
using System.Collections.Specialized;
using System.Reflection;
using System.Linq.Expressions;

namespace ORM05
{
    class DynamicMethodBuilder
    {
        private static readonly Dictionary<Type, Func<object, Dictionary<string, object>>> dictionaryCache = new Dictionary<Type, Func<object, Dictionary<string, object>>>();

        public static readonly Dictionary<Type, Delegate> typeTypeDeserializers = new Dictionary<Type, Delegate>();

        public static Func<DbFieldReader, T> GetTypeDeserializer<T>(EntityMapping entityMapping)
        {
            if (typeTypeDeserializers.ContainsKey(entityMapping.EntityType) == false)
            {
                ParameterExpression parExpr = Expression.Parameter(typeof(DbFieldReader), "r");


                var newExpr = Expression.New(entityMapping.EntityType.GetConstructors().First());

                var method = typeof(DbFieldReader).GetMethods().Where(c => c.Name == "GetValue" && c.IsGenericMethod).First();

                List<MemberBinding> memberBindings = new List<MemberBinding>();
                int index = 0;
                foreach (var item in entityMapping.Members)
                {
                    var callExpr = Expression.Call(parExpr, method.MakeGenericMethod(item.Member.PropertyType), Expression.Constant(index));
                    var memberAssignment = Expression.Bind(item.Member, callExpr);
                    memberBindings.Add(memberAssignment);
                    index++;
                }

                var init = Expression.MemberInit(newExpr, memberBindings);
                Expression<Func<DbFieldReader, User>> expr = (Expression<Func<DbFieldReader, User>>)Expression.Lambda(init, parExpr);

                var func = expr.Compile();
                typeTypeDeserializers.Add(entityMapping.EntityType, func);
            }
            return (Func<DbFieldReader, T>)typeTypeDeserializers[entityMapping.EntityType];
        }

        public static Dictionary<string, object> ConvertFromObject(object obj)
        {
            if (obj is Dictionary<string, object>)
                return (Dictionary<string, object>)obj;

            var nvc = obj as NameValueCollection;
            if (nvc != null)
            {
                Dictionary<string, object> dict = new Dictionary<string, object>(StringComparer.OrdinalIgnoreCase);
                foreach (string key in nvc.Keys)
                    dict.Add(key, nvc[key]);
                return dict;
            }

            var hs = obj as Hashtable;
            if (hs != null)
            {
                Dictionary<string, object> dict = new Dictionary<string, object>(StringComparer.OrdinalIgnoreCase);
                foreach (string key in hs.Keys.OfType<string>())
                    dict.Add(key, hs[key]);
                return dict;
            }
            var type = obj.GetType();
            lock (dictionaryCache)
            {
                Func<object, Dictionary<string, object>> getter;
                if (dictionaryCache.TryGetValue(type, out getter) == false)
                {
                    getter = CreateDictionaryGenerator(type);
                    dictionaryCache[type] = getter;
                }
                var dict = getter(obj);
                return dict;
            }
        }


        private static Func<object, Dictionary<string, object>> CreateDictionaryGenerator(Type type)
        {
            var dm = new DynamicMethod((string.Format("Dictionary{0}", Guid.NewGuid())), typeof(Dictionary<string, object>), new[] { typeof(object) }, type, true);
            ILGenerator il = dm.GetILGenerator();
            il.DeclareLocal(typeof(Dictionary<string, object>));
            il.Emit(OpCodes.Nop);
            //Dictionary<string, object> dic = new Dictionary<string, object>();
            il.Emit(OpCodes.Newobj, typeof(Dictionary<string, object>).GetConstructor(Type.EmptyTypes));
            il.Emit(OpCodes.Stloc_0);

            foreach (var item in type.GetProperties(BindingFlags.Public | BindingFlags.Instance))
            {
                string columnName = item.Name;
                il.Emit(OpCodes.Nop);
                il.Emit(OpCodes.Ldloc_0);
                il.Emit(OpCodes.Ldstr, columnName);
                il.Emit(OpCodes.Ldarg_0);
                il.Emit(OpCodes.Callvirt, item.GetGetMethod());
                if (item.PropertyType.IsValueType)
                {
                    il.Emit(OpCodes.Box, item.PropertyType);
                }
                il.Emit(OpCodes.Callvirt, typeof(Dictionary<string, object>).GetMethod("Add"));
            }
            il.Emit(OpCodes.Nop);

            il.Emit(OpCodes.Ldloc_0);
            il.Emit(OpCodes.Ret);
            var func = (Func<object, Dictionary<string, object>>)dm.CreateDelegate(typeof(Func<object, Dictionary<string, object>>));
            return func;
        }
    }
}
